#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import yaml

from zenml.core.pipelines.config.standard_config import GlobalKeys, SplitKeys

TEMPLATE_CONFIG = {
    GlobalKeys.VERSION: 1,
    GlobalKeys.SPLIT: {
        SplitKeys.CATEGORIZE_BY_: {'by': 'a_categorical_column',
                                   'ratio': {'train': 0.7,
                                             'eval': 0.3}},

        SplitKeys.INDEX_BY_: {'by': 'a_sequential_column',
                              'ratio': {'train': 0.6,
                                        'eval': 0.4}},
        SplitKeys.WHERE_: ['some_feature < 3',
                           'other_feature = 2']},
    GlobalKeys.FEATURES: {'training_feature_0': {},
                          'training_feature_1': {},
                          'training_feature_2': {},
                          'training_feature_3': {},
                          'training_feature_4': {}},

    GlobalKeys.EVALUATOR: {'evaluation_feature_0': {},
                           'evaluation_feature_1': {}},

    GlobalKeys.LABELS: {'label_column': {'loss': 'binary_crossentropy',
                                         'metrics': ['accuracy']}},

    GlobalKeys.TRAINER: {'architecture': 'feedforward',
                         'eval_batch_size': 64,
                         'last_activation': 'sigmoid',
                         'layers': [{'type': 'dense', 'units': 128},
                                    {'type': 'dense', 'units': 128},
                                    {'type': 'dense', 'units': 64}],
                         'num_output_units': 1,
                         'optimizer': 'adam',
                         'save_checkpoints_steps': 200,
                         'train_batch_size': 64,
                         'train_steps': 2000,
                         'type': 'classification'},

    GlobalKeys.PREPROCESSING: {
        'boolean': {'filling': [{'method': 'max',
                                 'parameters': {}}],
                    'label_tuning': [{'method': 'no_tuning',
                                      'parameters': {}}],
                    'resampling': [{'method': 'threshold',
                                    'parameters': {'c_value': 0,
                                                   'cond': 'greater',
                                                   'set_value': 1,
                                                   'threshold': 0}}],
                    'transform': [{'method': 'no_transform',
                                   'parameters': {}}]},
        'float': {'filling': [{'method': 'max',
                               'parameters': {}}],
                  'label_tuning': [{'method': 'no_tuning',
                                    'parameters': {}}],
                  'resampling': [{'method': 'mean',
                                  'parameters': {}}],
                  'transform': [{'method': 'scale_to_z_score',
                                 'parameters': {}}]},
        'integer': {'filling': [{'method': 'max', 'parameters': {}}],
                    'label_tuning': [{'method': 'no_tuning',
                                      'parameters': {}}],
                    'resampling': [{'method': 'mean',
                                    'parameters': {}}],
                    'transform': [{'method': 'scale_to_z_score',
                                   'parameters': {}}]},
        'string': {'filling': [{'method': 'custom',
                                'parameters': {'custom_value': ''}}],
                   'label_tuning': [{'method': 'no_tuning',
                                     'parameters': {}}],
                   'resampling': [{'method': 'mode',
                                   'parameters': {}}],
                   'transform': [{'method': 'compute_and_apply_vocabulary',
                                  'parameters': {}}]}}}

REQUIRED = [
    (GlobalKeys.VERSION,
     """
# This is the version of the configuration YAML!
#
# for more info: https://docs.maiot.io/your-pipeline-your-ideas/config-yaml
"""),
    (GlobalKeys.SPLIT, """
# The key 'split' is utilized to configure the process of splitting the 
dataset 
# into a training and an eval dataset.
#
#   - categorize_by:    A dictionary containing the options for a 
categorical split
#   - index_by:         A dictionary containing the options for an 
index-based split
#   - ratio:            A dictionary specifying a common percentage-based 
splitting
#                       approach for both categorical and index-based splits 
(if given)
#   - where:            A list of strings, which defines additional conditions
#                       while querying the datasource
#
# for more info: https://docs.maiot.io/splitting
"""),
    (GlobalKeys.FEATURES, """
# The main key `features` is used to define which set of features will be used 
# during the training and it allows the users to modify the pre-processing 
# steps filling, transform (and possibly resampling in case of sequential 
# data) for each one of these features.
#
# Structurally, each key under features represents a selected feature. 
# Under each key, the user has the chance to determine a method for each 
# pre-processing step to use. If it is not explicitly defined, the behaviour 
# will be inferred from the defaults based on the data type.
#
# for more info: https://docs.maiot.io/preprocessing/feature-selection
"""),
    (GlobalKeys.LABELS, """
# The main key labels is used to determine which data column will be used 
# as the label during the training. The inner structure of this column is 
# quite similar to the block features, where the keys denote the data columns  
# which are selected and the values include the pre-processing configuration
#
# for more info: https://docs.maiot.io/preprocessing/feature-selection
"""),
    (GlobalKeys.EVALUATOR, """
# The main key evaluator determines which data columns will be used in the 
# evaluation of the trained model. Structurally, it shares the same 
# structure as the features block.
#
# for more info: https://docs.maiot.io/preprocessing/feature-selection
"""),
    (GlobalKeys.TRAINER, """
# The main key trainer is used to configure the model and the training 
# parameters.
#
#   - architecture:             string value which determines the architecture 
#                               of the model, possible values include 
#                               'feedforward' for feedforward networks, 
#                               'sequence' for LSTM networks and 'sequence_ae' 
#                               for sequence-to-sequence autoencoders
#   - type:                     string value which defines the type of the 
#                               problem at hand, possible selections include 
#                               'regression', 'classification', 'autoencoder'
#   - train_batch_size:         the batch size during the training
#   - eval_batch_size:          the batch size during the eval steps
#   - train_steps:              the number of batches which should be 
#                               processed through the training
#   - save_checkpoints_steps:   the number of training batches, which will 
#                               indicate the frequency of the validation steps
#   - optimizer:                the name of the selected optimizer for the 
#                               training
#   - last_activation:          the name of the last layer in the model
#   - num_output_units:         the number of output units in the last layer
#   - sequence_length:          the length of the sequence in one data point 
#                               (provided only on a sequential problem setting)
#   - layers:                   list of dictionaries, which hold the layer 
#                               configurations
#
# for more info: https://docs.maiot.io/training-1/set-up-your-model-for
-training
"""),
    (GlobalKeys.PREPROCESSING, """
# This is the default preprocess block!
# 
# for more info: https://docs.maiot.io/preprocessing/data-transformations
"""),
]

OPTIONAL = [
    (GlobalKeys.TIMESERIES_, """
# This block configures the preprocessing steps specific to time-series 
# datasets. 
#
#   - resampling_rate_in_secs:      defines the resampling rate in seconds 
#                                   and it will be used at the corresponding 
#                                   pre-precessing step
#   - trip_gap_threshold_in_secs:   defines a maximum threshold in seconds in 
#                                   order to split the dataset into trips. 
#                                   Sequential transformations will occur once 
#                                   the data is split into trips based on 
this value.
#   - process_sequence_w_timestamp: specifies which data column holds the 
#                                   timestamp.
#   - process_sequence_w_category:  is an optional value, which, if provided, 
#                                   will be used to split the data into 
#                                   categories before the sequential processes
#   - sequence_shift:               defines the shift (in datapoints) while 
#                                   extracting sequences from the dataset
#
# for more info: https://docs.maiot.io/preprocessing/sequential-datasets
""")
]


def generate_comment_block(block, description):
    """
    Args:
        block:
        description:
    """
    return """##{filler}##
# {title} #
##{filler}##
{description}
""".format(filler='#' * len(block),
           title=block.upper(),
           description=description)


def generate_config_block(block, config):
    """
    Args:
        block:
        config:
    """
    return yaml.dump({block: config[block]}, default_flow_style=False) + '\n'


def save_pretty_yaml(config, output_path, no_docs):
    """
    Args:
        config:
        output_path:
        no_docs:
    """
    with open(output_path, "w") as output_file:
        for block, description in REQUIRED:
            if not no_docs:
                comment = generate_comment_block(block, description)
                output_file.writelines(comment)
            block = generate_config_block(block, config)
            output_file.writelines(block)

        for block, description in OPTIONAL:
            if block in config:
                if not no_docs:
                    comment = generate_comment_block(block, description)
                    output_file.writelines(comment)
                block = generate_config_block(block, config)
                output_file.writelines(block)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import apache_beam as beam
import tensorflow as tf
import json
import numpy as np
import base64
from typing import Dict, Text, Any, Tuple
from dateutil.parser import parse

from zenml.core.components.data_gen import constants
from zenml.core.components.data_gen.constants import BINARY_DATA, FILE_NAME, \
    BASE_PATH, FILE_PATH, FILE_EXT, METADATA, IMAGE, MASK
from zenml.core.components.data_gen.constants import DestinationKeys
from zenml.core.components.data_gen.types.interface import TypeInterface
from zenml.core.components.data_gen.utils import init_bq_table, DtypeInferrer
from zenml.core.components.data_gen.utils import is_smaller_than


@beam.ptransform_fn
@beam.typehints.with_input_types(beam.typehints.Dict[Text, Any])
@beam.typehints.with_output_types(beam.pvalue.PDone)
def WriteCVImagesToBQ(datapoints: Dict[Text, Any],
                      schema: Dict[Text, Any],
                      destination: Dict[Text, Any]) \
        -> beam.pvalue.PDone:
    """
    Args:
        datapoints:
        schema:
        destination:
    """
    allowed_ext = [".jpg", ".json", ".png", ".txt", ".jpeg"]

    cv_data = (datapoints
               | "FilterOutFiles" >> beam.Filter(lambda x:
                                                 x[FILE_EXT]
                                                 in allowed_ext)
               | "ConvertToNamedTuple" >> beam.Map(lambda x:
                                                   (x[FILE_NAME], x))
               | "CombineByName" >> beam.GroupByKey()
               # now we have a list of inputs grouped by file name URI
               # add and parse metadata, create TFExample
               | "AddTFExandMD" >> beam.Map(append_tf_example_to_image)
               )

    # Obtain the schema
    if schema:
        schema_dict = schema
        assert all(v in ['STRING', 'INTEGER', 'FLOAT', 'BYTES']
                   for _, v in schema.items())
        init_bq_table(
            destination[DestinationKeys.PROJECT],
            destination[DestinationKeys.DATASET],
            destination[DestinationKeys.TABLE],
            schema)
    else:
        schema = (cv_data
                  | 'InferSchema' >> beam.CombineGlobally(
                    DtypeInferrer(),
                    project=destination[DestinationKeys.PROJECT],
                    dataset=destination[DestinationKeys.DATASET],
                    table=destination[DestinationKeys.TABLE]))

        schema_dict = beam.pvalue.AsSingleton(schema)

    # filter out data points larger than 10 MiB (default)
    final_images = (cv_data
                    | "FilterBySize" >> beam.Filter(is_smaller_than)
                    )

    # TODO: For my dummy dataset, this fails with
    #  BrokenPipeError: [Errno 32] Broken pipe.
    #  SO hints at a possible row / request size error
    #  because the rows may be too big. More info:
    #  https://stackoverflow.com/questions/59856658/google-bigquery-payload
    #  -size-limit-of-10485760-bytes
    #  https://cloud.google.com/bigquery/quotas
    result = (final_images
              | "WriteToBQ" >> beam.io.WriteToBigQuery(
                table=destination[DestinationKeys.TABLE],
                dataset=destination[DestinationKeys.DATASET],
                project=destination[DestinationKeys.PROJECT],
                create_disposition="CREATE_NEVER",
                write_disposition='WRITE_EMPTY')
              )

    return result


def _int64_feature(value):
    """Returns an int64_list from a bool / enum / int / uint.

    Args:
        value:
    """
    input_value = value.flatten() if (hasattr(value, 'shape') and
                                      (len(value.shape) > 1)) else [value]
    return tf.train.Feature(int64_list=tf.train.Int64List(value=input_value))


def _float_feature(value):
    """Returns an int64_list from a bool / enum / int / uint.

    Args:
        value:
    """
    input_value = value.flatten() if (hasattr(value, 'shape') and
                                      (len(value.shape) > 1)) else [value]
    return tf.train.Feature(float_list=tf.train.FloatList(
        value=input_value))


def _bytes_feature(value):
    """Returns a bytes_list from a string / byte.

    Args:
        value:
    """
    if isinstance(value, type(tf.constant(0))):
        value = value.numpy()
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))


def _label_feature(value):
    """Returns an int64_list from a bool / enum / int / uint.

    Args:
        value:
    """
    # flatten first if it is a label array, which is the case e.g. in
    # computer vision + segmentation
    input_value = value.flatten() if (hasattr(value, 'shape') and
                                      (len(value.shape) > 1)) else [value]

    return tf.train.Feature(int64_list=tf.train.Int64List(
        value=input_value))


def _metadata_feature(metadata):
    # Returns all the metadata in the image as a string
    """
    Args:
        metadata:
    """
    metadata_as_bytes = str.encode(metadata)
    return tf.train.Feature(bytes_list=tf.train.BytesList(
        value=[metadata_as_bytes]))


def get_metadata(label_data: Text, img_filename: Text) -> Dict[Text, Any]:
    """
    Args:
        label_data (Text): Should be JSON-readable.
        img_filename (Text):

    Returns:
        Dict, additional metadata information: **metadata**
    """

    # TODO: This can potentially go bad if the JSON is too big
    data_dict = json.loads(label_data)

    try:
        metadata = data_dict.pop(METADATA)
    except KeyError:
        metadata = {}

    # JSON parsing puts empty metadata dict as the empty string,
    # so in that case coerce to dict just to be type safe
    if not metadata:
        metadata = {}

    return metadata


def parse_metadata(metadata_dict: Dict[Text, Any]):
    """
    Args:
        metadata_dict:
    """
    key_feature_map = {
        "int": _int64_feature,
        "float": _float_feature,
        "str": _metadata_feature,
        "bytes": _bytes_feature}

    md_features = {}
    new_metadata = metadata_dict.copy()  # copy of the data

    # separate check if value is list-like, also numpy.ndarray coercion
    for key, value in metadata_dict.items():
        val_type = type(value)
        if isinstance(value, list):
            value = np.array(value).flatten()
            val_type = type(value[0])

        # check whether a str is actually a timestamp
        if isinstance(value, str):
            try:
                ts = parse(value, fuzzy=True)
                ts = ts.strftime('%Y-%m-%dT%H:%M:%S.%f %Z')
                new_metadata[key] = ts
            except:
                pass

        try:
            md_features[key] = key_feature_map.get(val_type.__name__)(value)
        except KeyError:
            raise RuntimeError("Error: Unrecognized data type in TFExample "
                               "construction.")

    return new_metadata, md_features


def append_tf_example_to_image(data_tuple: Tuple[Text, Dict[Text, Any]]):
    """This function breaks down images, masks and metadata into a single dict
    that can be pushed into BigQuery.

    Args:
        data_tuple:

    Returns:
        Dict, with proper types, formatting ready for BQ insertion: **image_dict**
    """
    image_dict = {}
    feature = {}

    data_list = data_tuple[-1]
    # TODO: Check length of data_list to infer whether a label, image or
    #  mask is missing; proceed accordingly

    for data_dict in data_list:
        raw_data = data_dict.pop(BINARY_DATA)
        base_path = data_dict.pop(BASE_PATH)
        file_path = data_dict.pop(FILE_PATH)

        # image path, data is image data
        if "image" in file_path:
            image = tf.io.decode_image(raw_data).numpy()
            image_shape = image.shape

            feature.update({'height': _int64_feature(image_shape[0]),
                            'width': _int64_feature(image_shape[1]),
                            'depth': _int64_feature(image_shape[2]),
                            IMAGE: _bytes_feature(raw_data)})

        elif "mask" in file_path:
            mask = tf.io.decode_image(raw_data).numpy()
            mask_shape = mask.shape
            feature.update({'mask_height': _int64_feature(mask_shape[0]),
                            'mask_width': _int64_feature(mask_shape[1]),
                            'mask_depth': _int64_feature(mask_shape[2]),
                            MASK: _bytes_feature(raw_data)})

        elif "metadata" in file_path:
            metadata = get_metadata(raw_data, data_dict[FILE_NAME])
            parsed_metadata, md_feature = parse_metadata(metadata)
            # update feature dict with metadata information
            image_dict.update(parsed_metadata)
            feature.update(md_feature)

        else:
            continue

    tf_example = tf.train.Example(features=tf.train.Features(feature=feature))

    image_dict[constants.DATA_COL] = base64.b64encode(
        tf_example.SerializeToString())

    return image_dict


class CVImageConverter(TypeInterface):
    def convert(self):
        return WriteCVImagesToBQ



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Training pipeline step to create a pipeline that trains on data."""

import logging
from typing import Dict, Text, Any, List

from tfx.components.schema_gen.component import SchemaGen
from tfx.components.statistics_gen.component import StatisticsGen
from tfx.components.trainer.component import Trainer
from tfx.components.transform.component import Transform
from tfx.dsl.components.base import executor_spec
from tfx.extensions.google_cloud_ai_platform.trainer import \
    executor as ai_platform_trainer_executor
from tfx.proto import trainer_pb2

from zenml.core.backends.orchestrator.local.orchestrator_local_backend import \
    OrchestratorLocalBackend
from zenml.core.backends.processing.processing_local_backend import \
    ProcessingLocalBackend
from zenml.core.backends.training.training_local_backend import \
    TrainingLocalBackend
from zenml.core.components.data_gen.component import DataGen
from zenml.core.components.split_gen.component import SplitGen
from zenml.core.pipelines.base_pipeline import BasePipeline
from zenml.core.pipelines.utils import sanitize_name_for_ai_platform
from zenml.core.standards import standard_config as keys
from zenml.core.steps.deployer import DeployerStep
from zenml.core.steps.evaluator import EvaluatorStep
from zenml.core.steps.preprocesser.base_preprocesser import \
    BasePreprocesserStep
from zenml.core.steps.split.base_split_step import BaseSplitStep
from zenml.core.steps.timeseries import TimeseriesStep
from zenml.core.steps.trainer.base_trainer import BaseTrainerStep
from zenml.utils.enums import GDPComponent
from zenml.utils.post_training.post_training_utils import \
    get_statistics_artifact, \
    get_schema_artifact, get_pusher_artifact, \
    evaluate_single_pipeline, view_statistics, view_schema, detect_anomalies

logger = logging.getLogger(__name__)


class TrainingPipeline(BasePipeline):
    """Training Pipeline definition to

    TrainingPipeline is a general-purpose training pipeline for most ML
    training runs. One pipeline runs one experiment on a datasource.
    """
    PIPELINE_TYPE = 'training'

    def get_tfx_component_list(self, config: Dict[Text, Any]) -> List:
        specs = self.get_pipeline_spec(config)
        steps = config[keys.GlobalKeys.STEPS]

        component_list = []

        ############
        # RAW DATA #
        ############
        data_config = steps[keys.TrainingSteps.DATA]
        data = DataGen(
            source=data_config['source'],
            source_args=data_config['args']
        ).with_id(GDPComponent.DataGen.name)

        statistics_data = StatisticsGen(
            examples=data.outputs.examples
        ).with_id(GDPComponent.DataStatistics.name)

        schema_data = SchemaGen(
            statistics=statistics_data.outputs.output,
            infer_feature_shape=True,
        ).with_id(GDPComponent.DataSchema.name)

        component_list.extend([data,
                               statistics_data,
                               schema_data])

        datapoints = data.outputs.examples

        #################
        #   SPLITTING   #
        #################
        # Block to read the data from the corresponding BQ table
        split_config = steps[keys.TrainingSteps.SPLIT]
        splits = SplitGen(
            input_examples=datapoints,
            source=split_config['source'],
            source_args=split_config['args'],
            schema=schema_data.outputs.schema,
            statistics=statistics_data.outputs.output,
        ).with_id(GDPComponent.SplitGen.name)

        datapoints = splits.outputs.examples

        statistics_split = StatisticsGen(
            examples=datapoints
        ).with_id(GDPComponent.SplitStatistics.name)

        schema_split = SchemaGen(
            statistics=statistics_split.outputs.output
        ).with_id(GDPComponent.SplitSchema.name)

        schema = schema_split.outputs.schema

        component_list.extend([splits,
                               statistics_split,
                               schema_split])

        ##################
        # Data Validation #
        ##################
        from tfx.components.tuner.component import Tuner
        #################
        # PREPROCESSING #
        #################
        transform = Transform(
            preprocessing_fn='zenml.core.components.transform.transform_module'
                             '.preprocessing_fn',
            examples=datapoints,
            schema=schema,
            custom_config=steps[keys.TrainingSteps.PREPROCESSING]
        ).with_id(GDPComponent.Transform.name)

        component_list.extend([transform])

        ############
        # TRAINING #
        ############
        training_type = specs['training_args']['type']
        training_args = specs['training_args']['args']

        kwargs = {'custom_config': steps[keys.TrainingSteps.TRAINING]}

        if training_type == 'gcaip':
            kwargs.update({
                'custom_executor_spec': executor_spec.ExecutorClassSpec(
                    ai_platform_trainer_executor.GenericExecutor)})

            cloud_job_prefix = 'some_constant'  # TODO: Fix the constant issue

            from tfx.extensions.google_cloud_ai_platform.trainer.executor \
                import TRAINING_ARGS_KEY, JOB_ID_KEY
            kwargs['custom_config'].update(
                {TRAINING_ARGS_KEY: training_args,
                 JOB_ID_KEY: sanitize_name_for_ai_platform(cloud_job_prefix)})

        from tfx.components.trainer.executor import GenericExecutor
        Trainer.EXECUTOR_SPEC = executor_spec.ExecutorClassSpec(
            GenericExecutor)

        trainer = Trainer(
            transformed_examples=transform.outputs.transformed_examples,
            transform_graph=transform.outputs.transform_graph,
            run_fn='zenml.core.components.trainer.trainer_module.run_fn',
            schema=schema,
            train_args=trainer_pb2.TrainArgs(),
            eval_args=trainer_pb2.EvalArgs(),
            **kwargs
        ).with_id(GDPComponent.Trainer.name)

        component_list.extend([trainer])

        #############
        # EVALUATOR #
        #############
        # eval_step = EvaluatorStep.from_config(steps[keys.TrainingSteps.EVALUATION])
        # eval_config = eval_step.build_eval_config()
        # evaluator = Evaluator(
        #     examples=transform.outputs.transformed_examples,
        #     model_exports=trainer.outputs.model,
        #     eval_config=eval_config
        # ).with_id(GDPComponent.Evaluator.name)
        #
        # component_list.append(evaluator)

        ###########
        # SERVING #
        ###########

        # from tfx.components.pusher.component import Pusher
        # from tfx.extensions.google_cloud_ai_platform.pusher import \
        #     executor as ai_platform_pusher_executor
        # from tfx.proto import pusher_pb2
        #
        # # pusher
        # pusher_kwargs = {
        #     'model_export': trainer.outputs.output,
        #     'push_destination': pusher_pb2.PushDestination(
        #         filesystem=pusher_pb2.PushDestination.Filesystem(
        #             base_directory=spec['serving_model_dir']))
        # }
        #
        # if spec['serving_type'] == ServingTypes.gcaip.name:
        #     pusher_kwargs.update(
        #         {'custom_executor_spec': executor_spec.ExecutorClassSpec(
        #             ai_platform_pusher_executor.Executor)})
        # pusher_kwargs['custom_config'] = {
        #     'ai_platform_serving_args': spec['ai_platform_serving_args']
        # }
        # pusher_kwargs['instance_name'] = GDPComponent.Deployer.name
        #
        # pusher = Pusher(**pusher_kwargs)
        # component_list.append(pusher)

        return component_list

    def add_split(self, split_step: BaseSplitStep):
        self.steps_dict[keys.TrainingSteps.SPLIT] = split_step

    def add_timeseries(self, timeseries_step: TimeseriesStep):
        self.steps_dict[keys.TrainingSteps.SEQUENCING] = timeseries_step

    def add_preprocesser(self, preprocessor_step: BasePreprocesserStep):
        self.steps_dict[keys.TrainingSteps.PREPROCESSING] = preprocessor_step

    def add_trainer(self, trainer_step: BaseTrainerStep):
        self.steps_dict[keys.TrainingSteps.TRAINING] = trainer_step

    def add_evaluator(self, evaluator_step: EvaluatorStep):
        self.steps_dict[keys.TrainingSteps.EVALUATION] = evaluator_step

    def add_deployment(self, deployment_step: DeployerStep):
        self.steps_dict[keys.TrainingSteps.DEPLOYMENT] = deployment_step

    def view_statistics(self, magic: bool = False):
        """
        View statistics for training pipeline in HTML.

        Args:
            magic (bool): Creates HTML page if False, else
            creates a notebook cell.
        """
        uri = get_statistics_artifact(
            self.pipeline_name, GDPComponent.SplitStatistics.name)
        view_statistics(uri, magic)

    def view_schema(self):
        """View schema of data flowing in pipeline."""
        uri = get_schema_artifact(
            self.pipeline_name, GDPComponent.SplitSchema.name)
        view_schema(uri)

    def evaluate(self):
        """Evaluate pipeline."""
        return evaluate_single_pipeline(self.pipeline_name)

    def download_model(self, out_path: Text = None, overwrite: bool = False):
        """Download model to out_path"""
        model_path = get_pusher_artifact(self.pipeline_name)
        if out_path:
            from zenml.utils.path_utils import move
            move(model_path, out_path, overwrite)
        else:
            out_path = model_path
        logger.info(f'Model: {out_path}')

    def view_anomalies(self, split_name='eval'):
        """
        View schema of data flowing in pipeline.

        Args:
            split_name: name of split to detect anomalies on
        """
        stats_uri = get_statistics_artifact(
            self.pipeline_name, GDPComponent.SplitStatistics.name)
        schema_uri = get_schema_artifact(
            self.pipeline_name, GDPComponent.SplitSchema.name)
        detect_anomalies(stats_uri, schema_uri, split_name)

    def get_default_backends(self) -> Dict:
        """Gets list of default backends for this pipeline."""
        # For base class, orchestration is always necessary
        return {
            OrchestratorLocalBackend.BACKEND_KEY: OrchestratorLocalBackend(),
            ProcessingLocalBackend.BACKEND_KEY: ProcessingLocalBackend(),
            TrainingLocalBackend.BACKEND_KEY: TrainingLocalBackend(),
        }

    def steps_completed(self) -> bool:
        mandatory_steps = [keys.TrainingSteps.SPLIT,
                           keys.TrainingSteps.PREPROCESSING,
                           keys.TrainingSteps.TRAINING,
                           keys.TrainingSteps.DATA]
        for step_name in mandatory_steps:
            if step_name not in self.steps_dict.keys():
                raise AssertionError(f'Mandatory step {step_name} not added.')
        return True



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
from typing import List, Text

import tensorflow as tf
import tensorflow_transform as tft

from zenml.core.steps.base_step import BaseStep


class BaseTrainerStep(BaseStep):
    def __init__(self,
                 serving_model_dir: Text = None,
                 transform_output: Text = None,
                 train_files=None,
                 eval_files=None,
                 **unused_kwargs):
        """
        Constructor for BaseTrainerStep.

        Args:
            serving_model_dir:
            transform_output:
            train_files:
            eval_files:
            log_dir:
            schema:
        """
        self.serving_model_dir = serving_model_dir
        self.transform_output = transform_output
        self.train_files = train_files
        self.eval_files = eval_files
        self.schema = None
        self.log_dir = None

        # Infer schema and log_dir
        if self.transform_output is not None:
            tf_transform_output = tft.TFTransformOutput(self.transform_output)
            self.schema = tf_transform_output.transformed_feature_spec().copy()

        if self.serving_model_dir is not None:
            self.log_dir = os.path.join(
                os.path.dirname(self.serving_model_dir), 'logs')

    def get_run_fn(self):
        return self.run_fn

    def run_fn(self):
        pass

    def input_fn(self,
                 file_pattern: List[Text],
                 tf_transform_output: tft.TFTransformOutput):
        pass

    @staticmethod
    def _get_serve_tf_examples_fn(model, tf_transform_output):
        pass

    @staticmethod
    def _get_ce_eval_tf_examples_fn(model, tf_transform_output):
        pass

    @staticmethod
    def model_fn(train_dataset: tf.data.Dataset,
                 eval_dataset: tf.data.Dataset):
        pass



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Dict, Text, Any

from zenml.core.pipelines.config.standard_config import SplitKeys

TRAIN = 'train'
EVAL = 'eval'
TEST = 'test'
NOSPLIT = 'nosplit'
RATIO = 'ratio'
CATEGORIES = 'categories'
BY = 'by'


def check_split_ratio_sum(split_ratio_dict: Dict[Text, Any]):
    # using round because of float inaccuracies
    """
    Args:
        split_ratio_dict:
    """
    if round(sum(split_ratio_dict.values()), 15) != 1:
        raise AssertionError("Split ratios must sum up to 1.")


def check_split_ratio_keys(split_ratio_dict: Dict[Text, Any]):
    """Small helper to validate a split ratio dictionary by its keys. Allowed:
    Either train + eval, train + eval + test, or nosplit keys.

    Args:
        split_ratio_dict:
    """
    # Get fold names
    split_folds = list(split_ratio_dict.keys())
    num_folds = len(split_folds)

    required_folds = []

    if not split_folds:
        raise AssertionError("Error: The given split ratio "
                             "dictionary is empty.")

    # nosplit case
    if num_folds == 1:
        required_folds = [NOSPLIT]

    elif num_folds == 2:
        required_folds = [TRAIN, EVAL]

    elif num_folds == 3:
        required_folds = [TRAIN, EVAL, TEST]

    elif num_folds > 3:
        raise AssertionError(
            "Error: More than three fold keys are not supported at the "
            "moment.")

    if not required_folds:
        raise AssertionError(
            "Error: Something went wrong while validating the split ratio "
            "dict."
            "This should not happen. Contact a maiot GmbH engineer "
            "immediately!")

    if not all(str(fold) in required_folds for fold in split_folds):
        raise AssertionError(
            "Error while validating the given ratio dict. Only the following "
            "fold names are supported "
            "when specifying {0} folds: ".format(num_folds)
            + '[%s]' % ', '.join(required_folds))


def validate_split_ratio_dict(split_ratio_dict: Dict[Text, Any]):
    """
    Args:
        split_ratio_dict:
    """
    check_split_ratio_sum(split_ratio_dict)
    check_split_ratio_keys(split_ratio_dict)


def validate_split_config(split_config):
    """Helper to validate the supplied split config in terms of its entries.
    This is to prohibit some pathological cases (can be expanded).

    Args:
        split_config:
    """
    # ratio dict checks for vertical and horizontal split,
    # top-level ratio checks happen below
    for key in [SplitKeys.CATEGORIZE_BY_, SplitKeys.INDEX_BY_]:
        if key in split_config:
            split = split_config[key]
            if RATIO in split:
                ratio_dict = split[RATIO]
                validate_split_ratio_dict(ratio_dict)

    # top-level ratio dict exists
    if SplitKeys.RATIO_ in split_config:

        # keys in the top-level ratio dict
        top_level_ratio_dict = split_config[SplitKeys.RATIO_]

        # split ratio dict checks (sum + key consistency)
        validate_split_ratio_dict(top_level_ratio_dict)
        top_level_folds = list(top_level_ratio_dict.keys())

        # disallow all three ratio dicts to be set simultaneously
        if SplitKeys.CATEGORIZE_BY_ in split_config:
            # get the config for horizontal split
            categorize_by_cfg = split_config[SplitKeys.CATEGORIZE_BY_]
            if RATIO in categorize_by_cfg:
                category_ratio_dict = categorize_by_cfg[RATIO]

                categorize_by_folds = list(
                    category_ratio_dict.keys())
                # set comprehension to validate ratio dicts
                if not set(categorize_by_folds) == set(top_level_folds):
                    # Print the keys creating the error for verbosity
                    # to help the user
                    set_diff = (set(categorize_by_folds) - set(
                        top_level_folds)) \
                               or (set(top_level_folds) - set(
                        categorize_by_folds))
                    raise AssertionError(
                        "Error: Ratio dict inconsistency detected. Please "
                        "check that you "
                        "supply the same keys for each split ratio entry in "
                        "the config. Keys that do not appear in both ratio"
                        "dicts: {}".format(set_diff))

            if CATEGORIES in categorize_by_cfg:
                categories = categorize_by_cfg[CATEGORIES]
                # disallow category specification and top-level split at the
                # same time
                if isinstance(categories, dict):
                    raise AssertionError(
                        "Error: Specifying a top-level split ratio and a "
                        "category dict at the same time is "
                        "disallowed.")

            if SplitKeys.INDEX_BY_ in split_config:
                # get the config for vertical split
                index_by_cfg = split_config[SplitKeys.INDEX_BY_]

                # disallow all ratio dicts to be set simultaneously
                if all(RATIO in cfg for cfg in
                       [index_by_cfg, categorize_by_cfg]):
                    raise AssertionError(
                        "Error: Illegal ratio specification detected."
                        "Either drop the top-level \'ratio\' config key or "
                        "the lower-level index-by \'ratio\' key.")

        # disallow top-level ratio dict and index_by ratio dict to be set
        # simultaneously
        if SplitKeys.INDEX_BY_ in split_config:
            index_by_cfg = split_config[SplitKeys.INDEX_BY_]
            if RATIO in index_by_cfg:
                raise AssertionError("Illegal ratio specification detected. "
                                     "Specifying a top-level split ratio "
                                     "and an index-by ratio at the "
                                     "same time is not supported.")

    else:  # no top-level ratio dict
        if SplitKeys.CATEGORIZE_BY_ in split_config:
            # get the config for horizontal split
            categorize_by_cfg = split_config[SplitKeys.CATEGORIZE_BY_]
            if CATEGORIES in categorize_by_cfg:
                categories = categorize_by_cfg[CATEGORIES]
                if isinstance(categories, dict):
                    if RATIO in categorize_by_cfg:
                        # disallow category specification and category split at
                        # the same time
                        raise AssertionError(
                            "Error: Specifying a category split ratio and a "
                            "category dict "
                            "at the same time is disallowed.")

                    if NOSPLIT in categories:
                        # handle the case where nosplit is specified along with
                        # other options as an error
                        if len(categories.keys()) > 1:
                            raise AssertionError(
                                "Error: When specifying \'nosplit\', passing "
                                "other fold keys is disallowed.")

                elif isinstance(categories, list):
                    if RATIO not in categorize_by_cfg:
                        raise AssertionError(
                            "Error: When passing a list, the category split "
                            "ratio has to be specified,"
                            "either by a top-level or by a "
                            "category-level ratio dict.")

            else:
                # no top-level or category ratio keys specified
                if RATIO not in categorize_by_cfg:
                    raise AssertionError(
                        "Error: When passing a list, the category split ratio "
                        "has to be specified,"
                        "either by a top-level or "
                        "by a category-level ratio dict.")

            if SplitKeys.INDEX_BY_ in split_config:
                # get the config for vertical split
                index_by_cfg = split_config[SplitKeys.INDEX_BY_]
                if all(RATIO in cfg for cfg in
                       [index_by_cfg, categorize_by_cfg]):

                    category_ratio_dict = categorize_by_cfg[RATIO]
                    idx_ratio_dict = index_by_cfg[RATIO]

                    categorize_by_folds = list(
                        category_ratio_dict.keys())
                    index_by_folds = list(idx_ratio_dict.keys())
                    # set comprehension to validate ratio dicts
                    if not set(categorize_by_folds) == set(index_by_folds):
                        # Print the keys creating the error for verbosity
                        # to help the user
                        set_diff = (set(categorize_by_folds) - set(
                            index_by_folds)) or \
                                   (set(index_by_folds) - set(
                                       categorize_by_folds))
                        raise AssertionError(
                            "Error: Ratio dict inconsistency detected. Please "
                            "check that you supply the same keys for "
                            "each split ratio entry in the config. "
                            "Keys that do not appear in both ratio "
                            "dicts: {}".format(set_diff))

                if RATIO in index_by_cfg:
                    idx_ratio_dict = index_by_cfg[RATIO]

                    index_by_folds = list(idx_ratio_dict.keys())
                    if CATEGORIES in categorize_by_cfg:
                        categories = categorize_by_cfg[CATEGORIES]
                        if isinstance(categories, dict):
                            # set comprehension to validate category dict
                            # and index ratio dict
                            categorize_by_folds = list(categories.keys())
                            if not set(categorize_by_folds) == set(
                                    index_by_folds):
                                # Print the keys creating the error
                                # for verbosity to help the user
                                set_diff = (set(categorize_by_folds) - set(
                                    index_by_folds)) \
                                           or (set(index_by_folds) - set(
                                    categorize_by_folds))
                                raise AssertionError(
                                    "Error: Ratio dict inconsistency "
                                    "detected. Please check that you "
                                    "supply the same keys for each split"
                                    " ratio entry in the config. Keys that"
                                    " do not appear in both ratio "
                                    "dicts: {}".format(set_diff))
                        elif isinstance(categories, list):
                            if RATIO not in categorize_by_cfg:
                                raise AssertionError(
                                    "Error: When passing a list, the category"
                                    " split ratio has to be specified,"
                                    "either by a top-level or by a "
                                    "category-level ratio dict.")



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
from typing import Text, Dict, Any

import apache_beam as beam
from apache_beam.io import fileio
from zenml.core.components.data_gen.constants import LocalImageArgs
from zenml.core.components.data_gen.constants import BINARY_DATA, FILE_NAME, \
    BASE_PATH, FILE_PATH, FILE_EXT
from zenml.core.components.data_gen.sources.interface import SourceInterface


@beam.ptransform_fn
@beam.typehints.with_input_types(beam.Pipeline)
@beam.typehints.with_output_types(beam.typehints.Dict[Text, Any])
def ReadFilesFromDisk(pipeline: beam.Pipeline,
                      source_args: Dict[Text, Any]) -> beam.pvalue.PCollection:
    """
    Args:
        pipeline (beam.Pipeline):
        source_args:
    """
    base_path = source_args['base_path']

    wildcard_qualifier = "**"

    # we are ingesting all the files from the base path by
    # supplying the wildcard (hopefully?)
    file_pattern = os.path.join(base_path, wildcard_qualifier)

    files = (pipeline
             | fileio.MatchFiles(file_pattern)
             | fileio.ReadMatches()
             | beam.Reshuffle())

    files_and_contents = (files
                          | beam.Map(read_file_content, base_path)
                          )

    return files_and_contents


def read_file_content(file: beam.io.fileio.ReadableFile,
                      base_path: Text):
    """
    Args:
        file (beam.io.fileio.ReadableFile):
        base_path (Text):
    """
    data_dict = {FILE_PATH: file.metadata.path,
                 BINARY_DATA: file.read(),
                 BASE_PATH: base_path}

    base = os.path.basename(data_dict[FILE_PATH])
    data_dict[FILE_NAME] = base
    data_dict[FILE_EXT] = os.path.splitext(base)[1]

    return data_dict


class LocalImageSource(SourceInterface):
    ARG_KEYS = LocalImageArgs

    def beam_transform(self):
        return ReadFilesFromDisk



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import logging

from zenml.core.components.session_transform.sequence_transform import \
    SequenceTransform
from zenml.core.components.split_gen.component import SplitGen
from zenml.core.components.trainer.component import Trainer
from zenml.core.components.transform.component import Transform
from zenml.core.pipelines.config.standard_config import GlobalKeys


class ConfigLinter:

    def __init__(self, version: int = 1):
        """
        Args:
            version (int):
        """
        self.version = version

    @staticmethod
    def lint(config):
        """
        Args:
            config:
        """
        GlobalKeys.key_check(config)

        parsers = [SplitGen.lint,
                   Transform.lint,
                   Trainer.lint]

        if GlobalKeys.TIMESERIES_ in config:
            parsers.append(SequenceTransform.config_parser)

        for p_func in parsers:
            try:
                _ = p_func(config)
            except AssertionError as e:
                logging.error('Problem while reading the config '
                              'file: {}'.format(e))
                raise e
            except Exception as e:
                logging.error(e)
                raise AssertionError(
                    'We found an error in parsing your config YAML, but could '
                    'not infer what it is precisely. Please ensure your YAML '
                    'is formatted properly, and you are following the format '
                    'set in our docs: https://docs.maiot.io')



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Dict, Text, Any, List

import apache_beam as beam
from tfx import types
from tfx.dsl.components.base import base_executor
from tfx.components.example_gen.base_example_gen_executor import _WriteSplit
from tfx.types import artifact_utils

from zenml.core.components.data_gen.constants import DATA_SPLIT_NAME
from zenml.core.components.data_gen.constants import SpecParamKeys
from zenml.core.components.data_gen.utils import DtypeInferrer
from zenml.core.components.data_gen.utils import append_tf_example
from zenml.core.steps.data.base_data_step import BaseDataStep
from zenml.utils import source_utils


@beam.ptransform_fn
@beam.typehints.with_input_types(beam.typehints.Dict[Text, Any])
@beam.typehints.with_output_types(beam.pvalue.PDone)
def WriteToTFRecord(datapoints: Dict[Text, Any],
                    schema: Dict[Text, Any],
                    output_split_path: Text) -> beam.pvalue.PDone:
    """Infers schema and writes to TFRecord"""
    # Obtain the schema
    if schema:
        schema_dict = schema
        assert all(v in ['STRING', 'INTEGER', 'FLOAT']
                   for _, v in schema.items())
    else:
        schema = (datapoints
                  | 'InferSchema' >> beam.CombineGlobally(
                    DtypeInferrer()))
        schema_dict = beam.pvalue.AsSingleton(schema)

    return (datapoints
            | 'ToTFExample' >> beam.Map(append_tf_example, schema_dict)
            | 'WriteToTFRecord' >> _WriteSplit(
                output_split_path=output_split_path
            ))


class DataExecutor(base_executor.BaseExecutor):

    def Do(self,
           input_dict: Dict[Text, List[types.Artifact]],
           output_dict: Dict[Text, List[types.Artifact]],
           exec_properties: Dict[Text, Any]) -> None:
        """
        Args:
            input_dict:
            output_dict:
            exec_properties:
        """
        source = exec_properties[SpecParamKeys.SOURCE]
        args = exec_properties[SpecParamKeys.SOURCE_ARGS]

        c = source_utils.load_source_path_class(source)
        data_step: BaseDataStep = c(**args)

        # Get output split path
        examples_artifact = artifact_utils.get_single_instance(
            output_dict['examples'])
        split_names = [DATA_SPLIT_NAME]
        examples_artifact.split_names = artifact_utils.encode_split_names(
            split_names)
        output_split_path = artifact_utils.get_split_uri(
            [examples_artifact], DATA_SPLIT_NAME)

        with self._make_beam_pipeline() as p:
            (p
             | data_step.read_from_source()
             # | data_step.convert_to_dict()
             | WriteToTFRecord(data_step.schema, output_split_path))



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the identity split."""

from typing import Text, List, Any

from zenml.core.steps.split import constants
from zenml.core.steps.split.base_split_step import BaseSplitStep


def NoSplitPartitionFn(element: Any,
                       num_partitions: int) -> int:
    """
    Function for no split on data; to be used in a beam.Partition.
    Args:
        element: Data point, given as a tf.train.Example.
        num_partitions: Number of splits, unused here.

    Returns: An integer n, where 0 <= n <= num_partitions - 1.
    """
    return 0


class NoSplitStep(BaseSplitStep):

    def __init__(
            self,
            statistics=None,
            schema=None,
    ):
        """

        Args:
            statistics: Parsed statistics from a preceding StatisticsGen.
            schema: Parsed schema from a preceding SchemaGen.
        """
        super().__init__(statistics, schema)

    def partition_fn(self):
        return NoSplitPartitionFn, {}

    def get_split_names(self) -> List[Text]:
        return [constants.NOSPLIT]



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

# Test limit
LIMIT = 1000

# Split names and types
TRAIN = 'train'
EVAL = 'eval'
TEST = 'test'  # new test split name
NOSPLIT = 'nosplit'  # nosplit

DEFAULT_ROW_NUMBER = 'default_row_number'
DEFAULT_CATEGORY = 'default_category'

# Namings for execution params
BQ_PROJECT = 'bq_project'
BQ_DATASET = 'bq_dataset'
BQ_TABLE = 'bq_table'
TEST_MODE = 'test_mode'
WHERE = 'where'
FEATURES = 'features'
CAT_COL = 'cat_col'
CAT_TRAIN = 'cat_train'
CAT_EVAL = 'cat_eval'
CAT_TEST = 'cat_test'  # new test category
CAT_NOSPLIT = 'cat_nosplit'
IDX_COL = 'idx_col'

# index ratio constants for categorical and index based
C_RATIO_TRAIN = 'c_ratio_train'
C_RATIO_EVAL = 'c_ratio_eval'
C_RATIO_TEST = 'c_ratio_test'
C_RATIO_NOSPLIT = 'c_ratio_nosplit'
I_RATIO_TRAIN = 'idx_ratio_train'
I_RATIO_EVAL = 'idx_ratio_eval'
I_RATIO_TEST = 'idx_ratio_test'
I_RATIO_NOSPLIT = 'idx_ratio_nosplit'

# Various CATEGORIZE_BY and INDEX_BY sub-keys
RATIO = "ratio"
CATEGORIES = "categories"
BY = "by"
